// hierlb.cpp : implementation file
//

#include "stdafx.h"

#if !defined(HIERLB_H)
 #include "hierlb.h"
#endif

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#if defined(_DEBUG)
 IMPLEMENT_DYNAMIC(CHierListNode,CObject)
#endif

CHierListNode::CHierListNode( DWORD item, const CHierListNode* parent )
{
	m_ItemData  = item;
	m_pParent   = parent;
	m_bOpen     = FALSE;
	m_lastChild = (parent) ? parent->m_lastChild : 0; // inherit parents lines
	m_Level     = (parent) ? parent->m_Level+1 : 0;
}

/////////////////////////////////////////////////////////////////////////////
// CHierListBox

CHierListBox::CHierListBox( BOOL lines, WORD offset )
{
	m_bLines = lines;
	m_Offset = offset;
	m_bDisableLastChildCheck = FALSE;

	// you can enable drag and drop in your derived class
	// but it's up to you to implement insertion of items at the correct level.
	// InsertNode imposes some simple restrictions.	
	EnableDragDrop(FALSE);
}

CHierListBox::~CHierListBox()
{
}


BEGIN_MESSAGE_MAP(CHierListBox,CListBoxEx)
	//{{AFX_MSG_MAP(CHierListBox)
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

#define new DEBUG_NEW

/////////////////////////////////////////////////////////////////////////////
// CHierListBox message handlers


void CHierListBox::DeleteItem( LPDELETEITEMSTRUCT lpDIS )
{
	// This only deletes the CHierListNode*.  If you want to
	// delete your own data using this function, override it
	// and then call the base class version.
	CHierListNode* n = (CHierListNode*)lpDIS->itemData;
	ASSERT_NODE(n);
	
	// need to check last child state 
	// This function does not check if the item it's deleting is the
	// only remaining child of it's parent so the parent node will
	// remain open even though it has no children.
	// It also does not delete children of a deleted parent - that's the
	// job of CloseNode(int). DeleteNode(int) will close the node and then
	// delete it.
	if(!m_bDisableLastChildCheck)
	{
		if( n->m_Level != 0 )
		{
			if(n->GetLastChildBit())
			{
				// the deleted node was the last child so
				// check previous node
				// if same level set it's last child bit
				DWORD prevItem = GetItemData(lpDIS->itemID-1);
				if(prevItem != LB_ERR) 
				{
					CHierListNode* prevNode = (CHierListNode*)prevItem;
					ASSERT_NODE(prevNode);
					if(prevNode->m_Level == n->m_Level)
						prevNode->SetLastChildBit();
				}				
			}
		}
	}

	delete n;
}

/*
These are the possible ways of inserting a new node 

1) Insert at top of list box - level must be 0, parent must be NULL

2) Insert after another node - must be the same level and have the same parent as
	previous node.  The previous node must not be open.
	
This probably needs more sophisticated treatment.
*/
int CHierListBox::InsertNode( DWORD o, const CHierListNode* p, int index )
{
	if(p && p->m_Level>15)
	{
		AfxMessageBox( "This implementation supports\na maximum of 16 levels of indentation" );
		return LB_ERR;
	}
	
	CHierListNode* n = new CHierListNode(o,p);
	ASSERT(n);

	// Hierarchical list boxes can't have LBS_SORT so always insert
	if(n)
	{
		int insertedAt = InsertString( index, (LPCSTR)n );

		// last child check is disabled during internal operation OpenNode()
		// since the user is assumed to be inserting nodes correctly		
		if( !m_bDisableLastChildCheck && (insertedAt>=0) )
		{
			if(!insertedAt)
			{
				// Top of list box
				ASSERT( !n->m_pParent && !n->m_Level );
			}
			else
			{
				DWORD prevItem = GetItemData(insertedAt-1);
				CHierListNode* prev = (CHierListNode*)prevItem;
				ASSERT_NODE(prev);
				
				if(prevItem>=0)
				{
					ASSERT(n->m_Level==prev->m_Level);
					ASSERT(n->m_pParent==prev->m_pParent);
					ASSERT(!n->m_bOpen);
					if(prev->GetLastChildBit())
					{
						prev->ResetLastChildBit();
						n->SetLastChildBit();
					}
				}
			}
		}
		
		return insertedAt;
	}
	else
		return LB_ERRSPACE;
}			

BOOL CHierListBox::CloseNode( int index )
{
	// if node at index is open 
	//   find last descendant of this node
	// CloseNodes(a,b,c);

	DWORD item = GetItemData(index);
	if( item > 0 )
	{
		CHierListNode* n = (CHierListNode*)item;
		ASSERT_NODE(n);
		if(n->m_bOpen)
		{
			int lastchild = FindLastChild(index);
			if(lastchild!=LB_ERR)
			{
				CloseNodes(n->m_Level,lastchild,index);
				return TRUE;
			}
			else
				return FALSE;			
		}
		else
			return FALSE;		
	}
	else
		return FALSE;
}

BOOL CHierListBox::DeleteNode( int index )
{
	if( CloseNode( index ) )
	{
		DeleteString(index);
		return TRUE;
	}
	else	
		return FALSE;
}

BOOL CHierListBox::OpenNode( int index )
{
	long item1 = GetItemData(index);
	long item2 = GetItemData(index+1);

	if( item1!=LB_ERR && item2!=LB_ERR )
	{
		CHierListNode* n1 = (CHierListNode*)item1;
		CHierListNode* n2 = (CHierListNode*)item2;

		ASSERT_NODE(n1);
		ASSERT_NODE(n2);
		ASSERT(!n1->m_bOpen);
		ASSERT(n1->m_Level >= n2->m_Level);
		
		// node is closed so ask derived class to open
		int parentIndex = index;
		BeginWaitCursor();
		m_bDisableLastChildCheck = TRUE;
		n1->m_bOpen = OpenNode(n1,index);
		m_bDisableLastChildCheck = FALSE;
		EndWaitCursor();

		if(n1->m_bOpen)
		{
			SetRedraw(FALSE);

			// check that node was opened correctly
			DWORD nextItem = GetItemData(parentIndex+1);
			if( nextItem > 0 )
			{
				CHierListNode* nextNode = (CHierListNode*)nextItem;
				ASSERT_NODE(nextNode);
				if( nextNode->m_Level != n1->m_Level+1 )
					n1->m_bOpen = FALSE;
				else
				{
				#if defined(_DEBUG)
					int lastChild = FindLastChild( parentIndex );
					ASSERT( (CHierListNode*)GetItemData(parentIndex) == n1 );
					ASSERT(lastChild==index);
				#endif

					// set the last child bit for this node
					DWORD item = GetItemData(index);
					CHierListNode* n = (CHierListNode*)item;
					ASSERT_NODE(n);
					n->SetLastChildBit();

					MakeChildrenVisible(parentIndex,index);
				}
			}
			else
				n1->m_bOpen = FALSE;

			SetRedraw(TRUE);
			Invalidate();
		}
		return n1->m_bOpen;
	}
	else
	{
		// last item in list box?
		if( item1 >= 0 )
		{
			CHierListNode* n1 = (CHierListNode*)item1;
			int parentIndex = index;
			ASSERT_NODE(n1);

			SetRedraw(FALSE);

			BeginWaitCursor();
			m_bDisableLastChildCheck = TRUE;
			n1->m_bOpen = OpenNode(n1,index);
			m_bDisableLastChildCheck = FALSE;
			EndWaitCursor();
			// index of last child
#if defined(_DEBUG)
			int lastChild = FindLastChild(parentIndex);
			ASSERT(lastChild==index);
#endif
			// ok so set a bit in the lastChild mask in the lastChild item
			DWORD item = GetItemData(index);
			CHierListNode* n = (CHierListNode*)item;
			ASSERT_NODE(n);
			n->SetLastChildBit();

			MakeChildrenVisible(parentIndex,index);

			SetRedraw(TRUE);
			Invalidate();
			return n->m_bOpen;
		}
		else
			return FALSE;
	}
}

void CHierListBox::OnLButtonDblClk(UINT nFlags, CPoint point)
{
	int index = ItemFromPoint(point);

	if( index != LB_ERR )
	{
		long item = GetItemData(index);
		CHierListNode* n = (CHierListNode*)item;
		ASSERT_NODE(n);

		if(item<=0)
			return;
			
		if(n->m_bOpen)
			CloseNode(index);
		else
			OpenNode(index);
	}
}	

void CHierListBox::EnableLines(BOOL enable)
{
	if(enable ^ m_bLines)
	{
		m_bLines = enable;
		Invalidate();
	}
}

void CHierListBox::DrawLines( const CHierListNode* pNode, CDC* pDC, LPCRECT pRect )
{
	WORD mask     = pNode->GetLastChildMask();
	WORD maxLevel = pNode->m_Level;

	if(!maxLevel)
		return;

	// set line colour
	COLORREF textcol = m_pResources->ColorWindowText();
	pDC->SetBkColor(textcol);

	WORD levelmask = 1;

	// it looks better having the tees and corners at half bitmap height
	// than text height for large fonts.
	int bmHeight = m_pResources->BitmapHeight();
	
	for( WORD level=1; level<maxLevel+1; level++ )
	{
		int indent = m_Offset*(level-1) + m_Offset/2;

		BOOL bLastChild     = (BOOL)(levelmask & mask);
		BOOL bNextLastChild = (BOOL)((levelmask<<1) & mask);

		if(!bLastChild)
		{
			if(level == pNode->m_Level)
			{
				if( bNextLastChild )
				{
					// draw corner
					CRect r(indent,pRect->top,indent+1,pRect->top+bmHeight/2);
					pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r, NULL, 0, NULL );
					CRect r1(indent,pRect->top+bmHeight/2,indent+m_Offset/2,pRect->top+bmHeight/2+1);
					pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r1, NULL, 0, NULL );
				}
				else
				{
					// draw tee
					CRect r(indent,pRect->top,indent+1,pRect->bottom);
					pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r, NULL, 0, NULL );
					CRect r1(indent,pRect->top+bmHeight/2,indent+m_Offset/2,pRect->top+bmHeight/2+1);
					pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r1, NULL, 0, NULL );
				}
			}
			else
			{
				CRect r(indent,pRect->top,indent+1,pRect->bottom);
				pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r, NULL, 0, NULL );
			}
		}
		else
		{
			if(level == pNode->m_Level)
			{
				// draw corner
				CRect r(indent,pRect->top,indent+1,pRect->top+bmHeight/2);
				pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r, NULL, 0, NULL );
				CRect r1(indent,pRect->top+bmHeight/2,indent+m_Offset/2,pRect->top+bmHeight/2+1);
				pDC->ExtTextOut( indent, 0, ETO_OPAQUE, r1, NULL, 0, NULL );
			}
		}

		levelmask <<= 1;
	}
}

void CHierListBox::DrawItem(LPDRAWITEMSTRUCT lpDIS)
{
	CDC* pDC = CDC::FromHandle(lpDIS->hDC);

	// draw focus rectangle when no items in listbox
	if(lpDIS->itemID == (UINT)-1)
	{
		if(lpDIS->itemAction&ODA_FOCUS)
		{
			// rcItem.bottom seems to be 0 for LBS_OWNERDRAWVARIABLE list boxes 
			// so set it to something sensible
			lpDIS->rcItem.bottom = m_lfHeight;
			if(m_bLines)
				lpDIS->rcItem.left  += m_Offset;

			pDC->DrawFocusRect( &lpDIS->rcItem );
		}
		return;
	}
	else
	{
		ASSERT(m_pResources); //need to attach resources before creation/adding items

		CHierListNode* node = (CHierListNode*)lpDIS->itemData;
		ASSERT_NODE(node);

		int selChange   = lpDIS->itemAction&ODA_SELECT;
		int focusChange = lpDIS->itemAction&ODA_FOCUS;
		int drawEntire  = lpDIS->itemAction&ODA_DRAWENTIRE;

		int indent = GetLevelIndent(node->m_Level);

		lpDIS->rcItem.left += indent;

		if(selChange || drawEntire)
		{
			if(m_bLines)
				DrawLines( node, pDC, &lpDIS->rcItem );

			BOOL sel = lpDIS->itemState & ODS_SELECTED;

			COLORREF hlite   = ((sel)?(m_pResources->ColorHighlight()):(m_pResources->ColorWindow()));
			COLORREF textcol = ((sel)?(m_pResources->ColorHighlightText()):(m_pResources->ColorWindowText()));

			pDC->SetBkColor(hlite);
			pDC->SetTextColor(textcol);
			// fill the retangle with the background colour the fast way.
			pDC->ExtTextOut( lpDIS->rcItem.left, 0, ETO_OPAQUE, &lpDIS->rcItem, NULL, 0, NULL );

			// derived class can finish things off
 			CListBoxExDrawStruct ds( pDC, &lpDIS->rcItem, sel, (DWORD)node, lpDIS->itemID, m_pResources );
			DrawItemEx( ds );
		}

		if( focusChange || (drawEntire && (lpDIS->itemState&ODS_FOCUS)) )
			pDC->DrawFocusRect(&lpDIS->rcItem);
	}
}

int CHierListBox::FindLastChild(int parentIndex)
{
	// scan forwards from parent node 'n1' until end of listbox
	// or last child node is found - this should have the same index
	// 'index'.  Set this nodes last child flag

	// hopefully the index of the parent node is still the same
	DWORD parentItem  = GetItemData(parentIndex);
	if(parentItem==LB_ERR)
		return LB_ERR;
		
	CHierListNode* parentNode = (CHierListNode*)parentItem;

	BOOL bFoundLast = FALSE;
	int numItems    = GetCount();
	int lastChild   = 0;
	
	for( int i=parentIndex+1; (i<numItems) && !bFoundLast; i++ )
	{
		DWORD item = GetItemData(i);
		if( item > 0 )
		{
			CHierListNode* n = (CHierListNode*)item;
			ASSERT_NODE(n);
			if( n->m_Level < parentNode->m_Level+1 )
			{
				bFoundLast = TRUE;
				lastChild = i-1;
			}

		}
		else
		{
			ASSERT(0);
			return LB_ERR;
		}
	}

	if(!bFoundLast)
		lastChild = numItems-1;

	return lastChild;
}

void CHierListBox::MakeChildrenVisible(int parentIndex, int lastChildIndex)
{
	// need to scroll listbox if last item is not visible
	// if the last item will not fit then make sure parent is still visible

	if( lastChildIndex == LB_ERR )
		return;

	if( lastChildIndex == LB_ERRSPACE )
	{
		AfxMessageBox( "Too many items in list box" );
		return;
	}

	CRect clientrect;
	GetClientRect(clientrect);

	// find index of last item in list box
	int bottomIndex = ItemFromPoint( clientrect.BottomRight() - CSize(1,1) );

	if(( bottomIndex!=LB_ERR ) && ( lastChildIndex>bottomIndex ))
	{
		SetTopIndex(parentIndex);
	}
}

BOOL CHierListBox::CloseNodes( unsigned int level, int start, int end )
{
	SetRedraw(FALSE);

	BeginWaitCursor();

	m_bDisableLastChildCheck = TRUE;
		
	for( int i=start; i>end-1; i-- )
	{
		long item = GetItemData(i);
		ASSERT(item!=LB_ERR);
		if(item==LB_ERR)
			break;
		CHierListNode* n = (CHierListNode*)item;
		ASSERT_NODE(n);

		if( n->m_Level > level )
			DeleteString(i);
		else			
			if( n->m_Level == level )
				n->m_bOpen = FALSE;
	}

	m_bDisableLastChildCheck = FALSE;
	    
    EndWaitCursor();
    
	SetRedraw(TRUE);
	Invalidate();
	
	return TRUE;
}

BOOL CHierListBox::CloseToLevel( unsigned int level )
{
	// close all nodes up until 'level'
	return CloseNodes(level,GetCount()-1,0);
}


