#if !defined(HIERLB_H)
#define HIERLB_H

#if !defined(LSTBOXEX_H)
 #include "lstboxex.h"
#endif

// This class is used to hold an item in a Heirarchical list box

class CHierListNode 
#if defined(_DEBUG)
: public CObject
#endif
{
private:
	WORD  m_lastChild; // allows 16 levels of lines - use a DWORD for 32
public:
#if defined(_DEBUG)
	DECLARE_DYNAMIC(CHierListNode) // this really only needed for debug
#endif
public:
	DWORD m_ItemData; // untyped object contained by item
	WORD  m_Level;
	BOOL  m_bOpen;

	const CHierListNode* m_pParent;
	
	void ResetLastChildBit() { m_lastChild &= ~(0x01 << (m_Level-1)); }
	void SetLastChildBit()   { m_lastChild |=  (0x01 << (m_Level-1)); }
	BOOL GetLastChildBit() const { return m_lastChild & (0x01 << (m_Level-1)); }
	WORD GetLastChildMask() const { return m_lastChild; }
public:
	CHierListNode( DWORD item, const CHierListNode* );
};

#if defined(_DEBUG)
 inline void ASSERT_NODE(CHierListNode* n) 
 { 
 	ASSERT(n); 
 	ASSERT( DWORD(n) != LB_ERR );
 	ASSERT( DWORD(n) != LB_ERRSPACE );
 	ASSERT(n->IsKindOf(RUNTIME_CLASS(CHierListNode))); 
 }
#else
 inline void ASSERT_NODE(CHierListNode* n) { n=n; }
#endif

/////////////////////////////////////////////////////////////////////////////
// CHierListBox window

class CHierListNode;

class CHierListBox : public CListBoxEx
{
private:
	BOOL m_bLines; // do we want connecting lines
	WORD m_Offset;
	BOOL m_bDisableLastChildCheck;
private:
	void DrawLines(const CHierListNode*, CDC*, LPCRECT );
	BOOL CloseNodes(unsigned int, int, int);
	int  FindLastChild(int);
protected:
	// inherited from CListBox
	virtual void DrawItem( LPDRAWITEMSTRUCT );
	virtual int  CompareItem(LPCOMPAREITEMSTRUCT lpCIS) { return 0; }
		// delete memory allocated to hold node items.
	virtual void DeleteItem( LPDELETEITEMSTRUCT );

	// messages to derived class
	virtual BOOL OpenNode( const CHierListNode*, int& ) = 0;

	int  GetLevelIndent(int level) const { return m_Offset*level; }
	void MakeChildrenVisible(int,int);
// Construction
public:
	CHierListBox( BOOL=FALSE, WORD=16 );

// Attributes
public:

// Operations
public:
	BOOL CloseToLevel(unsigned int=0);
	BOOL CloseAll() { return CloseToLevel(); }
	BOOL CloseNode(int);
	int  InsertNode(DWORD, const CHierListNode*, int index=-1);
	BOOL DeleteNode(int);
	BOOL OpenNode(int);	

	void EnableLines(BOOL);
	BOOL LinesEnabled() const { return m_bLines; }
// Implementation
public:
	virtual ~CHierListBox();

	// Generated message map functions
protected:
	//{{AFX_MSG(CHierListBox)
	afx_msg void OnLButtonDblClk(UINT nFlags, CPoint point);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////
#endif // HIERLB_H
